@@ -268,9 +268,12 @@ static int test_invalid(void)
 	struct io_uring_cqe *cqe;
 	struct io_uring_sqe *sqe;
 
-	ret = t_create_ring(8, &ring, 0);
-	if (ret)
+	ret = t_create_ring(8, &ring, IORING_SETUP_SUBMIT_ALL);
+	if (ret) {
+		if (ret == -EINVAL)
+			return 0;
 		return ret;
+	}
 
 	ret = t_create_socket_pair(fds, true);
 	if (ret)
